"""
implementation of Pulay's DIIS algorithm to speed up iterative solution of
self charge consistent DFTB.

see http://vergil.chemistry.gatech.edu/notes/diis/node2.html

"""
from numpy import zeros, ravel, reshape, dot
from numpy.linalg import solve, norm, LinAlgError

class DIIS:
    def __init__(self, m):
        """
        Parameters:
        ===========
        m: maximum number of trial vectors that are kept in memory
        """
        self.m = m
        self.residual_vectors = []
        self.trial_vectors = []
    def nr_trial_vectors(self):
        return len(self.trial_vectors)
    def reset(self):
        self.residual_vectors = []
        self.trial_vectors = []        
    def add_trial_vector(self, p):
        """
        add the vector p to the set of iterative solutions
        """
        self.trial_vectors.append(p)
        if len(self.trial_vectors) >= 2:
            p_i = self.trial_vectors[-2]
            p_iplus1 = self.trial_vectors[-1]
            dp_i = p_iplus1 - p_i
            self.residual_vectors.append(dp_i)
        # throw away the oldes trial vectors
        if len(self.trial_vectors) > self.m:
            self.trial_vectors.pop(0)
            self.residual_vectors.pop(0)
    def get_DIIS_approximate(self):
        """
        determine the best linear combination of the previous
        solution vectors.

        Returns:
        ========
        p_(i+1), dp_(i+1)
        where 
        p_(i+1) optimized solution vector, next residual vector dp_(i+1)  

        We can test for convergence by calculating norm(dp_(i+1))/norm(p_(i+1))
        """
        n = len(self.residual_vectors)
        assert n > 1
        A = zeros((n+1,n+1))
        for i in range(0, n):
            for j in range(0, n):
                A[i,j] = dot(self.residual_vectors[i], self.residual_vectors[j])
            A[i,n] = -1.0
            A[n,i] = -1.0
        A[n,n] = 0.0
        b = zeros(n+1)
        b[-1] = -1.0
        # solve A*c = b
        c = solve(A,b)
        # p = sum_i c_i*p_i
        pDIIS = 0.0*self.trial_vectors[0]
        for i in range(0, n):
            pDIIS += c[i]*self.trial_vectors[i]
        # replace old trial vector by DIIS approximation
#        self.trial_vectors[-1] = pDIIS  # WRONG
        self.residual_vectors[-1] = pDIIS - self.trial_vectors[-2]
        dp = self.residual_vectors[-1]
        return pDIIS, dp
    def next_approximation(self, p):
        """
        add the trial vector p and replace it with the DIIS approximate

        Parameters:
        ===========
        p: numpy array of arbitrary shape

        Returns:
        ========
        p_next: numpy array with same shape as p
        """
        shape = p.shape
        pflat = ravel(p)
        self.add_trial_vector(pflat)
        if len(self.trial_vectors) < 3:
            return p
        try:
            pflat_next, dp = self.get_DIIS_approximate()
        except LinAlgError:
            print "DIIS failed: singular matrix!"
            pflat_next = 0.5*self.trial_vectors[-2] + 0.5*self.trial_vectors[-1]
            raise Exception("?")
        p_next = reshape(pflat_next, shape)
        return p_next
    def relative_change(self):
        """
        compute the relative change in the last iteration
          |p_(i+1) - p_i|/|p_i|
        """
        assert(self.m > 1)
        if len(self.residual_vectors) == 0:
            change = 1.0
        else:
            change = norm(self.residual_vectors[-1])/norm(self.trial_vectors[-1])
        return change
