"""
This module computes analytical gradients of eigenvalues and eigenvectors
for a hamiltonian that depends implicitly on the eigenvectors,
as is the case for the Hartree-Fock or Kohn-Sham hamiltonian.

The eigenvalue problem reads:

    H(X,p).X = S(p).X.E

where the columns of X contain the eigenvectors, E is a diagonal matrix
with the eigenenergies. H(X,p) is the hamiltonian and S(p) the overlap matrix,
both of which depend on an external parameter p (e.g. a nuclear coordinate).

The gradients dE/dp and dX/dp are found iteratively. The required input
consists of the partial gradients
  (partial H)/(partial p), (partial S)/(partial p)
and a function that given an approximation for dX/dp computes
  dHXp_mn = sum_ij dH_mn/dX_ij * dX_ij/dp.

For degenerate eigenvalues, the gradients may be discontinuous. The article
'Computation of eigenvalue and eigenvector derivatives for a general complex-valued
 eigensystem' by van der Aa, Ter Morsche and Mettheij
explains how to find continuous gradients.
"""
import numpy as np
import scipy.linalg as la

from DFTB import XYZ

def eigenvector_derivatives(gradH, gradS, dHXp, X, E):
    """
      find the derivatives of the eigenvalues E and eigenvectors X belonging
      to the generalized eigenvalue problem

        H X = S X E
        
      with respect to an external parameter

    Parameters:
    ===========
    gradH,gradS: gradients of the hamiltonian and overlap matrix, these are the partial
       derivatives w/r/t the external parameter
    dHXp: dH/dX dX/dp 
    X, E: eigenvectors and eigenvalues at reference position, solution of H.X=S.X.E
    """
    # compute eigenvectors and eigenvalues
    # of generalized eigenvalue problem
    nnuc,norb,norb = gradH.shape
    Xt = X.conjugate().transpose()
    # gradient of eigenvalues
    dEdp = np.zeros((nnuc,norb))
    # gradient of eigenvectors
    C = np.zeros((nnuc,norb,norb))
    dXdp = np.zeros((nnuc,norb,norb))
    for k in range(0, nnuc):
        assert la.norm(gradH[k,:,:] - gradH[k,:,:].conjugate().transpose()) < 1.0e-10
        assert la.norm(gradS[k,:,:] - gradS[k,:,:].conjugate().transpose()) < 1.0e-10
        for i in range(0, norb):
            dEdp[k,i] = np.dot(Xt[i,:], np.dot(gradH[k,:,:]+dHXp[k,:,:], X[:,i])) \
                       -np.dot(Xt[i,:], np.dot(gradS[k,:,:], X[:,i]*E[i]))
            # i == j
            C[k,i,i] = - 0.5 * np.dot(Xt[i,:], np.dot(gradS[k,:,:], X[:,i]))
            for j in range(0, norb):
                # i != j
                if i != j and abs(E[j]-E[i]) > 1.0e-6:
                    C[k,i,j] = (  np.dot(Xt[i,:], np.dot(gradH[k,:,:]+dHXp[k,:,:], X[:,j])) \
                                 -np.dot(Xt[i,:], np.dot(gradS[k,:,:], X[:,j]*E[j]))  ) \
                                   /(E[j]-E[i])
        dXdp[k,:,:] = np.dot(X, C[k,:,:])
        #
#        err = la.norm(np.dot(Xt, np.dot(gradS[k,:,:], X)) + (C[k,:,:] + C[k,:,:].transpose()))
#        assert err < 1.0e-5, "err = %s" % err
        #
    return dEdp, dXdp

def eigenvector_derivatives_iterative(gradH, gradS, dHXp_func, X, E, tol=1.0e-10, max_iter=500):
    """
    In case the Hamiltonian itself depends on the MO coefficients
    the equations for C, such that dX/dp = X.C have to be solved
    iteratively
    """
    # initial guess for gradients of MO coefficients
    nnuc,norb,norb = gradH.shape
    dXdp_last = np.zeros((nnuc,norb,norb))
    dEdp_last = np.zeros((nnuc,norb))
    # change of hamiltonian due to change of MO coefficients
    dHXp = dHXp_func(dXdp_last)
    for i in range(0, max_iter):
        dEdp, dXdp = eigenvector_derivatives(gradH, gradS, dHXp, X, E)
        change = la.norm(dXdp-dXdp_last) + la.norm(dEdp-dEdp_last)
        print "change: %s" % change
        if change < tol:
            print "MO gradients CONVERGED after %d iterations" % i
            break
        dHXp = dHXp_func(dXdp)
        dXdp_last = dXdp
        dEdp_last = dEdp
    else:
        raise RuntimeError("MO gradients not converged after %d iterations" % i)
    return dEdp, dXdp

################ TEST #####################

def H0(p):
    a11 = 2*np.cos(p)*(-2+2*p**2+(1+p**2)*np.cos(2*p)-4*p**2*np.sin(p))
    a12 = 8*p*(-2*np.cos(p)+np.sin(2*p))
    a13 = (-3+5*p**2)*np.cos(p)-(1+p**2)*np.cos(3*p)-4*p**2*np.sin(2*p)
    a21 = 4*p*(-2*np.cos(p)+np.sin(2*p))
    a22 = -2*np.cos(p)*(-5+3*p**2+8*np.sin(p))
    a23 = 4*p*(-2*np.cos(p)+np.sin(2*p))
    a31 = (-3+5*p**2)*np.cos(p)-(1+p**2)*np.cos(3*p)-4*p**2*np.sin(2*p)
    a32 = 8*p*(-2*np.cos(p)+np.sin(2*p))
    a33 = 2*np.cos(p)*(-2+2*p**2+(1+p**2)*np.cos(2*p)-4*p**2*np.sin(p))
    h = np.array(\
        [[a11,a12,a13],
         [a12,a22,a23],
         [a13,a23,a33]])
    h /= (2*(1+p**2))
    return h

def S(p):
    olp = 1.0/p
    s = np.array(\
        [[1.0,olp,olp],
         [olp,1.0,olp],
         [olp,olp,1.0]])
    return s
    
def G(p):
    olp = 1.0/p
    g = np.array(\
        [[1.0,olp,0.0],
         [olp,1.0,0.0],
         [0.0,0.0,0.0]])
    return g

def H(X,p):
    g = G(p)
    h = H0(p) + np.dot(X.transpose(), np.dot(g,X))
    return h

def gradients(X,p):
    dhdp = (H(X,p+dp)-H(X,p))/dp
    dsdp = (S(p+dp)-S(p))/dp
    g = G(p)
    m,n = g.shape
    dhdp = np.reshape(dhdp, (1,m,n))
    dsdp = np.reshape(dsdp, (1,m,n))
    def dHXp_func(dXdp):
        dHXp = np.dot(dXdp[0,:,:].T, np.dot(g, X)) + np.dot(X.T, np.dot(g,dXdp[0,:,:]))
        dHXp = np.reshape(dHXp, (1,m,n))
        return dHXp
    return dsdp, dhdp, dHXp_func

def solve_self_consistently(p):
    s = S(p)
    X_last = np.zeros(s.shape)
    for i in range(0, 100):
        h = H(X_last,p)
        E,X = la.eigh(h,s)
        change = la.norm(X-X_last)
        print "change: %s" % change
        if change < 1.0e-10:
            print "CONVERGED"
            break
        X_last = X
    return E,X

def eigenvector_derivatives_numeric(p):
    E0,X0 = solve_self_consistently(p)
    E1,X1 = solve_self_consistently(p+dp)
    dE = (E1-E0)/dp
    dX = (X1-X0)/dp
    return dE, dX

def test_eigenvector_derivative():
    p = 1.57

    E,X = solve_self_consistently(p)
    print "Analytic"
    gradS, gradH, dHXp_func = gradients(X,p)
    dE_an, dX_an = eigenvector_derivatives_iterative(gradH, gradS, dHXp_func, X, E)    
    print "Numerical"
    dE_num, dX_num = eigenvector_derivatives_numeric(p)
    print "numerical eigenvalue gradients"
    print dE_num
    print "analytical eigenvalue gradients"
    print dE_an
    print "numerical eigenvector gradients"
    print dX_num
    print "analytical eigenvector gradients"
    print dX_an
    err_dE = la.norm(dE_num-dE_an)
    err_dX = la.norm(dX_num-dX_an)
    assert err_dX < 1.0e-4, "err(dX) = %s" % err_dX
    assert err_dE < 1.0e-4, "err(dE) = %s" % err_dE

############ DFTB TEST ##################
def dftb_numerical_gradients(dftb2, atomlist):
    dftb2.setGeometry(atomlist)

    dftb2.getEnergy()
    orbe0 = dftb2.getKSEnergies()
    orbs0 = dftb2.getKSCoefficients()
    #
    atomlist0 = dftb2.getGeometry()
    vec = XYZ.atomlist2vector(atomlist0)
    Nat = len(atomlist)
    Norb,Norb = orbs0.shape
    dorbe = np.zeros((3*Nat, Norb))
    dorbs = np.zeros((3*Nat, Norb, Norb))
    for i in range(0, 3*Nat):
        vec_dp = np.copy(vec)
        vec_dp[i] += dp
        atomlist = XYZ.vector2atomlist(vec_dp, atomlist0)
        dftb2.setGeometry(atomlist)

        dftb2.getEnergy()
        orbe1 = dftb2.getKSEnergies()
        orbs1 = dftb2.getKSCoefficients()
    
        dorbe[i,:] = (orbe1-orbe0)/dp
        dorbs[i,:,:] = (orbs1-orbs0)/dp

    return dorbe, dorbs


def test_dftb_eigenvector_derivative():
    from DFTB.XYZ import read_xyz, extract_keywords_xyz
    from DFTB.DFTB2 import DFTB2
    from DFTB import utils
    import sys
    import os.path
    import string
    usage = "Usage: %s <xyz-file>\n" % sys.argv[0]
    usage += "   --help option will provide more information\n"
    parser = utils.OptionParserFuncWrapper([DFTB2.__init__], usage)
    (options, args) = parser.parse_args()
    if len(args) < 1:
        print usage
        exit(-1)
    xyz_file = args[0]
    kwds = extract_keywords_xyz(xyz_file)
    # read first structure to get the atom types
    structures = read_xyz(xyz_file)
    atomlist = structures[0]

    options["verbose"] = 0
    dftb2 = DFTB2(atomlist, **options)
    # disable long-range correction
    dftb2.long_range_correction = 0

    dftb2.setGeometry(atomlist)

    dftb2.getEnergy()
    gradS, gradH, dHXp_func = dftb2.gradients_HandS()

    E = dftb2.getKSEnergies()
    X = dftb2.getKSCoefficients()
    dE_an, dX_an = eigenvector_derivatives_iterative(gradH, gradS, dHXp_func, X, E)    
    
    dE_num, dX_num = dftb_numerical_gradients(dftb2, atomlist)
    print "eigenvalues"
    print E
    print "numerical eigenvalue gradients"
    print dE_num
    print "analytical eigenvalue gradients"
    print dE_an
#    print "numerical eigenvector gradients"
#    print dX_num
#    print "analytical eigenvector gradients"
#    print dX_an
    err_dE = la.norm(dE_num-dE_an)
    err_dX = la.norm(dX_num-dX_an)
    assert err_dE < 1.0e-4, "err(dE) = %s" % err_dE
    assert err_dX < 1.0e-4, "err(dX) = %s" % err_dX


if __name__ == "__main__":
    dp = 0.000001
#    test_eigenvector_derivative()
    test_dftb_eigenvector_derivative()
