"""
gradients of KS orbital energies
and MO coefficients
"""
import numpy as np
import scipy.linalg as la

from DFTB import XYZ

def eigenvector_derivatives(H, S, gradH, gradS, dHXp, X, E):
    """
      find the derivatives of the eigenvalues E and eigenvectors X belonging
      to the generalized eigenvalue problem

        H X = S X E
        
      with respect to an external parameter

    Parameters:
    ===========
    H, S: hamiltonian and overlap matrix at the reference position
    gradH,gradS: gradients of the hamiltonian and overlap matrix, these are the partial
       derivatives w/r/t the external parameter
    dHXp: dH/dX dX/dp 
    X, E: eigenvectors and eigenvalues at reference position, solution of H.X=S.X.E
    """
    # compute eigenvectors and eigenvalues
    # of generalized eigenvalue problem
    nnuc,norb,norb = gradH.shape
    Xt = X.conjugate().transpose()
    ndim = len(H)
    # gradient of eigenvalues
    dEdp = np.zeros((nnuc,norb))
    # gradient of eigenvectors
    C = np.zeros((nnuc,norb,norb))
    dXdp = np.zeros((nnuc,norb,norb))
    for k in range(0, nnuc):
        assert la.norm(gradH[k,:,:] - gradH[k,:,:].conjugate().transpose()) < 1.0e-10
        assert la.norm(gradS[k,:,:] - gradS[k,:,:].conjugate().transpose()) < 1.0e-10
        for i in range(0, norb):
            dEdp[k,i] = np.dot(Xt[i,:], np.dot(gradH[k,:,:]+dHXp[k,:,:], X[:,i])) \
                       -np.dot(Xt[i,:], np.dot(gradS[k,:,:], X[:,i]*E[i]))
            # i == j
            C[k,i,i] = - 0.5 * np.dot(Xt[i,:], np.dot(gradS[k,:,:], X[:,i]))
            for j in range(0, norb):
                # i != j
                if i != j and abs(E[j]-E[i]) > 1.0e-6:
                    C[k,i,j] = (  np.dot(Xt[i,:], np.dot(gradH[k,:,:]+dHXp[k,:,:], X[:,j])) \
                                 -np.dot(Xt[i,:], np.dot(gradS[k,:,:], X[:,j]*E[j]))  ) \
                                   /(E[j]-E[i])
        dXdp[k,:,:] = np.dot(X, C[k,:,:])
        #
        err = la.norm(np.dot(Xt, np.dot(gradS[k,:,:], X)) + (C[k,:,:] + C[k,:,:].transpose()))
        assert err < 1.0e-5, "err = %s" % err
        #
    return dEdp, dXdp

################ TEST #####################
def numerical_gradients(dftb2, atomlist):
    dftb2.setGeometry(atomlist)
#    dftb2.runNonSCC()
    dftb2.getEnergy()
    orbe0 = dftb2.getKSEnergies()
    orbs0 = dftb2.getKSCoefficients()
    #
    atomlist0 = dftb2.getGeometry()
    dp = 0.0001
    vec = XYZ.atomlist2vector(atomlist0)
    Nat = len(atomlist)
    Norb,Norb = orbs0.shape
    dorbe = np.zeros((3*Nat, Norb))
    dorbs = np.zeros((3*Nat, Norb, Norb))
    for i in range(0, 3*Nat):
        vec_dp = np.copy(vec)
        vec_dp[i] += dp
        atomlist = XYZ.vector2atomlist(vec_dp, atomlist0)
        dftb2.setGeometry(atomlist)
#        dftb2.runNonSCC()
        dftb2.getEnergy()
        orbe1 = dftb2.getKSEnergies()
        orbs1 = dftb2.getKSCoefficients()
    
        dorbe[i,:] = (orbe1-orbe0)/dp
        dorbs[i,:,:] = (orbs1-orbs0)/dp

    return dorbe, dorbs

if __name__ == "__main__":
    from DFTB.XYZ import read_xyz, extract_keywords_xyz
    from DFTB.DFTB2 import DFTB2
    from DFTB import utils
    import sys
    import os.path
    import string
    usage = "Usage: %s <xyz-file>\n" % sys.argv[0]
    usage += "   --help option will provide more information\n"
    parser = utils.OptionParserFuncWrapper([DFTB2.__init__], usage)
    (options, args) = parser.parse_args()
    if len(args) < 1:
        print usage
        exit(-1)
    xyz_file = args[0]
    kwds = extract_keywords_xyz(xyz_file)
    # read first structure to get the atom types
    structures = read_xyz(xyz_file)
    atomlist = structures[0]

    dftb2 = DFTB2(atomlist, **options)
    dftb2.setGeometry(atomlist)
#    dftb2.runNonSCC()
    dftb2.getEnergy()
    # gradients
#    S, H = dftb2.getH0andS()
    S, H = dftb2.getHandS()
#    gradS, gradH = dftb2.gradients_H0andS()
#    dHXp = 0.0*gradH
    gradS, gradH, gradH_X = dftb2.gradients_HandS()
    
#    print "gradS"
#    print gradS[3,:,:]
#    print "gradH0"
#    print gradH0[3,:,:]
    
    E = dftb2.getKSEnergies()
    X = dftb2.getKSCoefficients()
    grad_orbs_last = 0.0*gradH
    dHXp = gradH_X(grad_orbs_last)
    for i in range(0, 100):
        grad_orbe, grad_orbs = eigenvector_derivatives(H, S, gradH, gradS, dHXp, X, E)
        change = la.norm(grad_orbs-grad_orbs_last)
        print "grad orbe"
        print grad_orbe
        if change < 1.0e-10:
            print "MO gradients CONVERGED after %d iterations" % i
            break
        grad_orbs_last = grad_orbs
        dHXp = gradH_X(grad_orbs)
        print "change: %s" % change

    grad_orbe_num, grad_orbs_num = numerical_gradients(dftb2, atomlist)
    print "analytical gradient of orbital energies"
    print grad_orbe
    print "numerical gradient of orbital energies"
    print grad_orbe_num

    err = la.norm(abs(grad_orbe - grad_orbe_num))
    assert err < 1.0e-1, "error of grad_orbe = %s" % err
    print "analytical gradient of MO coefficients"
    print grad_orbs[1,:,:]
    print "numerical gradient of MO coefficients"
    print grad_orbs_num[1,:,:]
    err = la.norm(abs(grad_orbs - grad_orbs_num))
    assert err < 1.0e-2, "error of grad_orbs = %s" % err
