#!/usr/bin/env python
"""
Create geometries for a scan along one internal degree of freedom
"""
import MolecularCoords
from DFTB import XYZ
from DFTB import AtomicData

import string
from os.path import expandvars, expanduser, basename

import numpy as np

if __name__ == "__main__":
    import sys
    import optparse
    usage = "Usage: %s <xyz file> <name of varying degree of freedom>\n" % basename(sys.argv[0])
    usage += "  The scan variable can be given either by its name (e.g. 'r(c0-c1)') or by its index in the z-matrix (e.g. 0).\n"
    usage += "  Angles are given in degrees and distances in Angstrom.\n"

    parser = optparse.OptionParser(usage)
    parser.add_option("--scan_range", dest="scan_range", help="Tuple with (icmin, icmax, Npts) giving the range over which the internal coordinate is varied. [default: %default]", default="(1.0, 2.0, 10)")
    parser.add_option("--out_xyz", dest="out_xyz", help="save the geometries of the scan to this xyz-file [default: %default]", default="scan.xyz") 
    parser.add_option("--out_var", dest="out_var", help="save values of internal coordinate along scan to a table [default: %default]", default="scan_var.dat")

    (opts,args) = parser.parse_args()
    if len(args) < 2:
        print usage
        exit(-1)

    xyz_file = args[0]
    scan_var = args[1]

    atomlist = XYZ.read_xyz(xyz_file)[0]
    zmat0 = MolecularCoords.cartesian2Zmatrix(atomlist)
    labels = MolecularCoords.Zmatrix_labels(atomlist)

    print "                Z-Matrix:"
    print "                ========="
    print "Index  Coordinate                     Value"
    print "---------------------------------------------------"
    for ic,(icname,icval) in enumerate(zip(labels,zmat0)):
        icnamestr = string.ljust(icname, 30)
        if icname[0] in ["a", "d"]:
            # convert angles from radians to degrees
            icval *= 180.0/np.pi
            icvalstr = string.rjust("%3.7f deg" % icval, 15)
            print "%2.1d   %s %s" % (ic, icnamestr, icvalstr)           
        elif icname[0] == "r":
            icval *= AtomicData.bohr_to_angs
            icvalstr = string.rjust("%3.7f Ang" % icval, 15)
            print "%2.1d   %s %s" % (ic, icnamestr, icvalstr)
    try:
        ic_scan = int(scan_var)
    except ValueError:
        try:
            ic_scan = labels.index(scan_var)
        except ValueError:
            print "%s does not seem to be an internal variable in the Z-matrix! Maybe you need to reorder the atoms in the xyz-file." % scan_var
            exit(-1)

    ic_label = labels[ic_scan]
    (icmin, icmax, Npts) = eval(opts.scan_range)
    scan_range = np.linspace(icmin, icmax, Npts)
    scan_range_deg_ang = np.copy(scan_range)

    title = "Scan of the variable %s from %s to %s in %s points" % (ic_label, icmin, icmax, Npts)
    print title

    if ic_label[0] in ["a", "d"]:
        # convert back to radians
        scan_range *= np.pi/180.0
    elif ic_label[0] == "r":
        # convert back to bohr
        scan_range /= AtomicData.bohr_to_angs

    scan_atomlists = []
    for icval in scan_range:
        zmat = np.copy(zmat0)
        zmat[ic_scan] = icval
        vec = MolecularCoords.Zmatrix2cartesian(zmat)
        scan_atomlists.append( XYZ.vector2atomlist(vec, atomlist) )

    XYZ.write_xyz(expanduser(expandvars(opts.out_xyz)), scan_atomlists, title=title)
    print "Scan geometries saved to %s" % opts.out_xyz
    np.savetxt(expanduser(expandvars(opts.out_var)), scan_range_deg_ang)
    print "Values of scan variables saved to %s" % opts.out_var

