#!/usr/bin/env python
"""
Optimizes all geometries in a xyz-file 
"""

from DFTB.PES import PotentialEnergySurfaces
from DFTB import XYZ, utils, AtomicData
from DFTB.Timer import GlobalTimer as T

import numpy as np
import numpy.linalg as la

if __name__ == "__main__":
    import sys
    from scipy import optimize

    if len(sys.argv) < 3:
        print "Usage: %s <xyz-file> I" % sys.argv[0]
        print "  This script optimizes all geometry in the xyz-file on the I-th electronic state\n"
        print "  All geometries should contain the same atoms in the same order, but may have different\n"
        print "  connectivities.\n"
        print "  type --help to see all options"
        print "  to reduce the amount of output add the option --verbose=0"
        exit(-1)
    
    #
    xyz_file = sys.argv[1]    # path to xyz-file
    I = int(sys.argv[2])      # index of electronic state
    # Read the geometries from the xyz-file
    geometries = XYZ.read_xyz(xyz_file)
    atomlist_ref = geometries[0]
    # read the charge of the molecule from the comment line in the xyz-file
    kwds = XYZ.extract_keywords_xyz(xyz_file)

    # FIND ENERGY MINIMUM
    # f is the objective function that should be minimized
    # it returns (f(x), f'(x))
    def f(x):
        if I == 0 and type(pes.tddftb.XmY) != type(None):
            # only ground state is needed. However, at the start
            # a single TD-DFT calculation is performed to initialize
            # all variables (e.g. X-Y), so that the program does not
            # complain about non-existing variables.
            enI, gradI = pes.getEnergyAndGradient_S0(x)
        else:
            energies, gradI = pes.getEnergiesAndGradient(x, I)
            enI = energies[I]
        print "E = %2.7f     |grad| = %2.7f" % (enI, la.norm(gradI))
        return enI, gradI

    # This function appends the current optimized geometry to the output xyz-file.
    def save_xyz(x, mode="a"):
        atomlist_opt = XYZ.vector2atomlist(x, atomlist_ref)
        XYZ.write_xyz(xyz_opt, [atomlist_opt], \
                      title="charge=%s" % kwds.get("charge",0),\
                      mode=mode)
        
    #
    options = {'gtol': 1.0e-7, 'norm': 2}

    xyz_opt = xyz_file.replace(".xyz", "_opt.xyz")

    for i,atomlist in enumerate(geometries[-100:]):
        print "optimizing geometry %d" % i
        # initialize the TD-DFTB calculator
        pes = PotentialEnergySurfaces(atomlist, Nst=max(I+1,2), **kwds)
        # convert geometry to a vector
        x0 = XYZ.atomlist2vector(atomlist)
        # optimize the geometry with BFGS algorithm
        res = optimize.minimize(f, x0, method="BFGS", jac=True, options=options)
        # save optimized geometry
        xopt = res.x
        if i == 0:
            mode = "w"
        else:
            mode = "a"
        save_xyz(xopt, mode=mode)
        print "optimized geometry %d appended to %s" % (i, xyz_opt)

    # timing
    print T
    
